#include "Mesh.h"
#include <float.h>
#include <windows.h>
#include <gl/gl.h>
#include "Vec3.h"
#include <map>

//////////////////////
// Vertex Declarations
//////////////////////

float &Vertex::operator[] ( int ind )
{
	assert ( ind >= 0 && ind < 8 );
	return data [ ind ];
}

float *Vertex::getPos ( void )
{
	return &(data[5]);
}

float *Vertex::getNormal ( void )
{
	return &(data [ 2 ]);
}

float *Vertex::getTex ( void )
{
	return &(data [ 0 ]);
}


////////////////////////
// Triangle Declarations
////////////////////////
int &Triangle::operator[] ( int ind )
{
	assert ( ind >= 0 && ind < 3 );
	return data [ ind ];
}

class VertexKey
{
public:
	int key [ 3 ];
};

// comparison operator for map
bool operator < ( const VertexKey &k1, const VertexKey &k2 )
{
	for ( int i = 0; i < 3; i++ )
	{
		if ( k1.key [ i ] < k2.key [ i ] )
		{
			return true;
		}
		if ( k1.key [ i ] > k2.key [ i ] )
		{
			return false;
		}
	}

	return false;
}

////////////////////
// Mesh Declarations
////////////////////
#include "glInfo.h"
#include "glExt.h"
PFNGLGENBUFFERSARBPROC pglGenBuffersARB = 0;                     // VBO Name Generation Procedure
PFNGLBINDBUFFERARBPROC pglBindBufferARB = 0;                     // VBO Bind Procedure
PFNGLBUFFERDATAARBPROC pglBufferDataARB = 0;                     // VBO Data Loading Procedure
PFNGLBUFFERSUBDATAARBPROC pglBufferSubDataARB = 0;               // VBO Sub Data Loading Procedure
PFNGLDELETEBUFFERSARBPROC pglDeleteBuffersARB = 0;               // VBO Deletion Procedure
PFNGLGETBUFFERPARAMETERIVARBPROC pglGetBufferParameterivARB = 0; // return various parameters of VBO
PFNGLMAPBUFFERARBPROC pglMapBufferARB = 0;                       // map VBO procedure
PFNGLUNMAPBUFFERARBPROC pglUnmapBufferARB = 0;                   // unmap VBO procedure
#define glGenBuffersARB           pglGenBuffersARB
#define glBindBufferARB           pglBindBufferARB
#define glBufferDataARB           pglBufferDataARB
#define glBufferSubDataARB        pglBufferSubDataARB
#define glDeleteBuffersARB        pglDeleteBuffersARB
#define glGetBufferParameterivARB pglGetBufferParameterivARB
#define glMapBufferARB            pglMapBufferARB
#define glUnmapBufferARB          pglUnmapBufferARB

Mesh::Mesh ( void )
{ 
	textureID = -1; 
	vbos [ 0 ] = vbos [ 1 ] = -1; 
	strcpy ( drawNames [ 0 ], "Immediate Mode" );
	strcpy ( drawNames [ 1 ], "Compute Normal Immediate Mode" );
	strcpy ( drawNames [ 2 ], "Display List" );
	strcpy ( drawNames [ 3 ], "Array Pointers" );
	strcpy ( drawNames [ 4 ], "Interleaved Arrays" );
	strcpy ( drawNames [ 5 ], "Vertex Buffer Objects" );
}

Mesh *Mesh::loadObj ( char *filename )
{
	FILE *fptr = fopen ( filename, "rt" );
	vector<Vec3> verts;
	vector<Vec3> normals;
	vector<Vec3> texCoords;
	char buffer [ 1024 ];
	Vec3 temp;
	Mesh *rvalue = new Mesh ( );
	map<VertexKey, int> vertexMap;
	bool hasNormals = false;

	if ( fptr == NULL )
	{
		return NULL;
	}

	while ( fgets ( buffer, 1024, fptr ) != NULL )
	{
		char *token = strtok ( buffer, " \t" );

		switch ( token [ 0 ] )
		{
		case 'v': // vertex, normal or texture coordinate
			if ( stricmp ( "v", token ) == 0 )
			{
				token = strtok ( NULL, " \t" );
				temp [ 0 ] = atof ( token );
				token = strtok ( NULL, " \t" );
				temp [ 1 ] = atof ( token );
				token = strtok ( NULL, " \t" );
				temp [ 2 ] = atof ( token );

				verts.push_back ( temp );
			}
			else if ( stricmp ( "vn", token ) == 0 )
			{
				token = strtok ( NULL, " \t" );
				temp [ 0 ] = atof ( token );
				token = strtok ( NULL, " \t" );
				temp [ 1 ] = atof ( token );
				token = strtok ( NULL, " \t" );
				temp [ 2 ] = atof ( token );

				normals.push_back ( temp );
			}
			else if ( stricmp ( "vt", token ) == 0 )
			{
				token = strtok ( NULL, " \t" );
				temp [ 0 ] = atof ( token );
				token = strtok ( NULL, " \t" );
				temp [ 1 ] = atof ( token );
				temp [ 2 ] = 0;

				texCoords.push_back ( temp );
			}
			else
			{
				fprintf ( stderr, "Unrecognized token: %s\n", token );
			}
			break;
		case 'f': // face
			// tokenize vertices first... then tokenize each vertex
			{
				vector<char *> vertTokens;
				vector<int> tempFace;

				token = strtok ( NULL, " \t" );
				while ( token != NULL )
				{
					if ( token [ 0 ] == '#' ) // comment for rest of line... break
					{
						break;
					}
					vertTokens.push_back ( token );
					token = strtok ( NULL, " \t" );
				}
				for ( int i = 0; i < vertTokens.size ( ); i++ )
				{
					VertexKey k;
					k.key [ 0 ] = k.key [ 1 ] = k.key [ 2 ] = -1;
					token = strpbrk ( vertTokens [ i ], "/" );
					for ( int j = 0; j < 3; j++ )
					{
						if ( token != NULL )
						{
							token [ 0 ] = NULL;
						}
						// could be the first character and should be skipped
						if ( token != vertTokens [ i ] )
						{
							k.key [ j ] = atoi ( vertTokens [ i ] );
							if ( k.key [ j ] < 0 )
							{
								// negative indexing
								if ( j == 0 )
								{
									k.key [ j ] = verts.size ( ) + k.key [ j ] + 1;
								}
								else if ( j == 1 )
								{
									k.key [ j ] = texCoords.size ( ) + k.key [ j ] + 1;
								}
								else
								{
									k.key [ j ] = normals.size ( ) + k.key [ j ] + 1;
								}
							}
						}
						if ( token == NULL )
						{
							break;
						}
						vertTokens [ i ] = token + 1;
						token = strpbrk ( vertTokens [ i ], "/" );
					}

					// check if it's the first time we've seen this vertex
					map<VertexKey, int>::iterator it = vertexMap.find ( k );
					if ( it == vertexMap.end ( ) )
					{
						// it is the first time, so add the vertex to the mesh
						Vertex v;
						v.getPos ( ) [ 0 ] = verts [ k.key [ 0 ] - 1 ] [ 0 ];
						v.getPos ( ) [ 1 ] = verts [ k.key [ 0 ] - 1 ] [ 1 ];
						v.getPos ( ) [ 2 ] = verts [ k.key [ 0 ] - 1 ] [ 2 ];
						if ( k.key [ 1 ] > 0 )
						{
							v.getTex ( ) [ 0 ] = texCoords [ k.key [ 1 ] - 1 ] [ 0 ];
							v.getTex ( ) [ 1 ] = texCoords [ k.key [ 1 ] - 1 ] [ 1 ];
						}
						if ( k.key [ 2 ] > 0 )
						{
							hasNormals = true;
							v.getNormal ( ) [ 0 ] = normals [ k.key [ 2 ] - 1 ] [ 0 ];
							v.getNormal ( ) [ 1 ] = normals [ k.key [ 2 ] - 1 ] [ 1 ];
							v.getNormal ( ) [ 2 ] = normals [ k.key [ 2 ] - 1 ] [ 2 ];
						}

						rvalue->verts.push_back ( v );
						vertexMap [ k ] = rvalue->verts.size ( ) - 1;
						tempFace.push_back ( rvalue->verts.size ( ) - 1 );
					}
					else
					{
						tempFace.push_back ( it->second );
					}
				}

				// now we must triangulate the face
				Triangle tri;
				tri [ 0 ] = tempFace [ 0 ];
				for ( int j = 1; j < tempFace.size ( ) - 1; j++ )
				{
					tri [ 1 ] = tempFace [ j ];
					tri [ 2 ] = tempFace [ j + 1 ];

					rvalue->tris.push_back ( tri );
				}
			}
			break;
		case '#': // comment... ignore
		case 's': // smoothing ... ignore
		case 'g': // group... ignore
		case 'u': // usemtl ... ignore
			break;
		default:
			fprintf ( stderr, "Unrecognized token: %s\n", token );
			break;
		}
	}
	fclose ( fptr );

	if ( !hasNormals )
	{
		rvalue->computeNormals ( );
	}

	return rvalue;
}

void Mesh::draw ( void )
{
	int i;

	glBegin ( GL_TRIANGLES );
		for ( i = 0; i < tris.size ( ); i++ )
		{
			glTexCoord2fv ( verts [ tris [ i ] [ 0 ] ].getTex ( ) );
			glNormal3fv ( verts [ tris [ i ] [ 0 ] ].getNormal ( ) );
			glVertex3fv ( verts [ tris [ i ] [ 0 ] ].getPos ( ) );

			glTexCoord2fv ( verts [ tris [ i ] [ 1 ] ].getTex ( ) );
			glNormal3fv ( verts [ tris [ i ] [ 1 ] ].getNormal ( ) );
			glVertex3fv ( verts [ tris [ i ] [ 1 ] ].getPos ( ) );

			glTexCoord2fv ( verts [ tris [ i ] [ 2 ] ].getTex ( ) );
			glNormal3fv ( verts [ tris [ i ] [ 2 ] ].getNormal ( ) );
			glVertex3fv ( verts [ tris [ i ] [ 2 ] ].getPos ( ) );
		}
	glEnd ( );
}

void Mesh::drawFlat ( void )
{
	int i;

	glBegin ( GL_TRIANGLES );
		for ( i = 0; i < tris.size ( ); i++ )
		{
			Vec3 *v1 = (Vec3 *)verts [ tris [ i ] [ 0 ] ].getPos ( );
			Vec3 *v2 = (Vec3 *)verts [ tris [ i ] [ 1 ] ].getPos ( );
			Vec3 *v3 = (Vec3 *)verts [ tris [ i ] [ 2 ] ].getPos ( );
			Vec3 norm = Vec3::crossProd ( *v2 - *v1, *v3 - *v1 );
			norm.normalize ( );

			glNormal3fv ( &(norm [ 0 ]) );
			glTexCoord2fv ( verts [ tris [ i ] [ 0 ] ].getTex ( ) );
			glVertex3fv ( verts [ tris [ i ] [ 0 ] ].getPos ( ) );

			glNormal3fv ( &(norm [ 0 ]) );
			glTexCoord2fv ( verts [ tris [ i ] [ 1 ] ].getTex ( ) );
			glVertex3fv ( verts [ tris [ i ] [ 1 ] ].getPos ( ) );

			glNormal3fv ( &(norm [ 0 ]) );
			glTexCoord2fv ( verts [ tris [ i ] [ 2 ] ].getTex ( ) );
			glVertex3fv ( verts [ tris [ i ] [ 2 ] ].getPos ( ) );
		}
	glEnd ( );
}

void Mesh::drawDisplayList ( void )
{
	static int listNum = -1;

	if ( listNum == -1 )
	{
		listNum = glGenLists ( 1 );
		glNewList ( listNum, GL_COMPILE_AND_EXECUTE );
			draw ( );
		glEndList ( );
	}
	else
	{
		glCallList ( listNum );
	}
}

void Mesh::drawInterleavedArrays ( void )
{
	glEnableClientState ( GL_VERTEX_ARRAY );
	glEnableClientState ( GL_TEXTURE_COORD_ARRAY );
	glEnableClientState ( GL_NORMAL_ARRAY );

	glInterleavedArrays ( GL_T2F_N3F_V3F, 0, &(verts [ 0 ] [ 0 ] ) );
	glDrawElements ( GL_TRIANGLES, tris.size ( ) * 3, GL_UNSIGNED_INT, &(tris [ 0 ] [ 0 ]) );

	glDisableClientState ( GL_VERTEX_ARRAY );
	glDisableClientState ( GL_TEXTURE_COORD_ARRAY );
	glDisableClientState ( GL_NORMAL_ARRAY );
}

void Mesh::drawArrayPointers ( void )
{
	glEnableClientState ( GL_VERTEX_ARRAY );
	glEnableClientState ( GL_TEXTURE_COORD_ARRAY );
	glEnableClientState ( GL_NORMAL_ARRAY );

	glVertexPointer ( 3, GL_FLOAT, 8 * sizeof ( float ), verts [ 0 ].getPos ( ) );
	glTexCoordPointer ( 2, GL_FLOAT, 8 * sizeof ( float ), verts [ 0 ].getTex ( ) );
	glNormalPointer ( GL_FLOAT, 8 * sizeof ( float ), verts [ 0 ].getNormal ( ) );

	glDrawElements ( GL_TRIANGLES, tris.size ( ) * 3, GL_UNSIGNED_INT, &(tris [ 0 ] [ 0 ]) );

	glDisableClientState ( GL_VERTEX_ARRAY );
	glDisableClientState ( GL_TEXTURE_COORD_ARRAY );
	glDisableClientState ( GL_NORMAL_ARRAY );
}

bool Mesh::initVBOs ( void )
{
    glInfo glInfo;
    glInfo.getInfo();

    // check VBO is supported by your video card
    if(glInfo.isExtensionSupported("GL_ARB_vertex_buffer_object"))
    {
        // get pointers to GL functions
        glGenBuffersARB = (PFNGLGENBUFFERSARBPROC)wglGetProcAddress("glGenBuffersARB");
        glBindBufferARB = (PFNGLBINDBUFFERARBPROC)wglGetProcAddress("glBindBufferARB");
        glBufferDataARB = (PFNGLBUFFERDATAARBPROC)wglGetProcAddress("glBufferDataARB");
        glBufferSubDataARB = (PFNGLBUFFERSUBDATAARBPROC)wglGetProcAddress("glBufferSubDataARB");
        glDeleteBuffersARB = (PFNGLDELETEBUFFERSARBPROC)wglGetProcAddress("glDeleteBuffersARB");
        glGetBufferParameterivARB = (PFNGLGETBUFFERPARAMETERIVARBPROC)wglGetProcAddress("glGetBufferParameterivARB");
        glMapBufferARB = (PFNGLMAPBUFFERARBPROC)wglGetProcAddress("glMapBufferARB");
        glUnmapBufferARB = (PFNGLUNMAPBUFFERARBPROC)wglGetProcAddress("glUnmapBufferARB");

        // check once again VBO extension
        if(glGenBuffersARB && glBindBufferARB && glBufferDataARB && glBufferSubDataARB &&
           glMapBufferARB && glUnmapBufferARB && glDeleteBuffersARB && glGetBufferParameterivARB)
        {
			// good
        }
        else
        {
			return false;
        }
    }
    else
    {
		return false;
    }

	if ( vbos [ 0 ] != -1 )
	{
		glDeleteBuffersARB ( 2, vbos );
	}

    glGenBuffersARB(2, vbos);
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbos [ 0 ]);
    glBufferDataARB(GL_ARRAY_BUFFER_ARB, sizeof ( float ) * 8 * verts.size ( ), &(verts [ 0 ] [ 0 ]), GL_STATIC_DRAW_ARB);
    glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, vbos [ 1 ]);
    glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB, sizeof ( int ) * 3 * tris.size ( ), &(tris [ 0 ] [ 0 ]), GL_STATIC_DRAW_ARB);

	glBindBufferARB ( GL_ARRAY_BUFFER_ARB, vbos [ 0 ] );
	glInterleavedArrays ( GL_T2F_N3F_V3F, 0, NULL );

	return true;
}

void Mesh::drawVBO ( void )
{
	glEnableClientState ( GL_VERTEX_ARRAY );
	glEnableClientState ( GL_TEXTURE_COORD_ARRAY );
	glEnableClientState ( GL_NORMAL_ARRAY );

	glBindBufferARB ( GL_ELEMENT_ARRAY_BUFFER_ARB, vbos [ 1 ] );
	glDrawElements ( GL_TRIANGLES, tris.size ( ) * 3, GL_UNSIGNED_INT, NULL );

	glDisableClientState ( GL_VERTEX_ARRAY );
	glDisableClientState ( GL_TEXTURE_COORD_ARRAY );
	glDisableClientState ( GL_NORMAL_ARRAY );
}

void Mesh::calcBoundingBox ( Vec3 &minCoord, Vec3 &maxCoord )
{
	minCoord [ 0 ] = minCoord [ 1 ] = minCoord [ 2 ] = FLT_MAX;
	maxCoord [ 0 ] = maxCoord [ 1 ] = maxCoord [ 2 ] = -FLT_MAX;

	for ( int i = 0; i < verts.size ( ); i++ )
	{
		for ( int j = 0; j < 3; j++ )
		{
			if ( verts [ i ].getPos ( ) [ j ] < minCoord [ j ] )
			{
				minCoord [ j ] = verts [ i ].getPos ( ) [ j ];
			}
			if ( verts [ i ].getPos ( ) [ j ] > maxCoord [ j ] )
			{
				maxCoord [ j ] = verts [ i ].getPos ( ) [ j ];
			}
		}
	}
}

void Mesh::translateSelf ( Vec3 &vec )
{
	for ( int i = 0; i < verts.size ( ); i++ )
	{
		verts [ i ].getPos ( ) [ 0 ] += vec [ 0 ];
		verts [ i ].getPos ( ) [ 1 ] += vec [ 1 ];
		verts [ i ].getPos ( ) [ 2 ] += vec [ 2 ];
	}
}

void Mesh::scaleSelf ( float s )
{
	Vec3 scale;
	scale [ 0 ] = scale [ 1 ] = scale [ 2 ] = s;

	scaleSelf ( scale );
}

void Mesh::scaleSelf ( Vec3 &s )
{
	for ( int i = 0; i < verts.size ( ); i++ )
	{
		verts [ i ].getPos ( ) [ 0 ] *= s [ 0 ];
		verts [ i ].getPos ( ) [ 1 ] *= s [ 1 ];
		verts [ i ].getPos ( ) [ 2 ] *= s [ 2 ];
	}
}

void Mesh::computeNormals ( void )
{
	int i;

	for ( i = 0; i < verts.size ( ); i++ )
	{
		verts [ i ].getNormal ( ) [ 0 ] = 0;
		verts [ i ].getNormal ( ) [ 1 ] = 0;
		verts [ i ].getNormal ( ) [ 2 ] = 0;
	}

	for ( i = 0; i < tris.size ( ); i++ )
	{
		Vec3 *v1 = (Vec3 *)(verts [ tris [ i ] [ 0 ] ].getPos ( ));
		Vec3 *v2 = (Vec3 *)(verts [ tris [ i ] [ 1 ] ].getPos ( ));
		Vec3 *v3 = (Vec3 *)(verts [ tris [ i ] [ 2 ] ].getPos ( ));

		Vec3 norm = Vec3::crossProd ( *v2 - *v1, *v3 - *v1 );

		v1 = (Vec3 *)(verts [ tris [ i ] [ 0 ] ].getNormal ( ));
		v2 = (Vec3 *)(verts [ tris [ i ] [ 1 ] ].getNormal ( ));
		v3 = (Vec3 *)(verts [ tris [ i ] [ 2 ] ].getNormal ( ));

		*v1 += norm;
		*v2 += norm;
		*v3 += norm;
	}

	for ( i = 0; i < verts.size ( ); i++ )
	{
		Vec3 *v = (Vec3 *)(verts [ i ].getNormal ( ));

		v->normalize ( );
	}
}