#ifndef MESH_H
#define MESH_H

#include <vector>
#include <assert.h>
#include "Vec3.h"
#include <string.h>

using namespace std;

class Vertex
{
public:
	float data [ 8 ]; // u, v, nx, ny, nz, x, y, z

	float &operator[] ( int ind );

	float *getPos ( void );

	float *getNormal ( void );

	float *getTex ( void );
};

class Triangle
{
public:
	int data [ 3 ];

	int &operator[] ( int ind );
};

class Mesh
{
protected:
	vector<Vertex> verts;
	vector<Triangle> tris;
	int textureID;
	unsigned int vbos [ 2 ];

public:
	char drawNames [ 6 ] [ 256 ];

public:
	Mesh ( void );

	static Mesh *loadObj ( char *filename );

	int getNumVerts ( void ) { return verts.size ( ); }
	int getNumTris ( void ) { return tris.size ( ); }

	void drawMesh ( int type )
	{
		switch ( type )
		{
		case 0:
			draw ( );
			break;
		case 1:
			drawFlat ( );
			break;
		case 2:
			drawDisplayList ( );
			break;
		case 3:
			drawArrayPointers ( );
			break;
		case 4:
			drawInterleavedArrays ( );
			break;
		case 5:
			drawVBO ( );
			break;
		}
	}

	void draw ( void );
	void drawFlat ( void );
	void drawDisplayList ( void );
	void drawArrayPointers ( void );
	void drawInterleavedArrays ( void );
//	void drawArrayElement ( void );
	bool initVBOs ( void );
	void drawVBO ( void );

	void calcBoundingBox ( Vec3 &minCoord, Vec3 &maxCoord );

	void translateSelf ( Vec3 &vec );

	void scaleSelf ( float s );
	void scaleSelf ( Vec3 &s );

	// WARNING: destroys any normals already in vertices
	void computeNormals ( void );
};

#endif