#include "PerspectiveView.h"
#include <gl/gl.h>
#include <gl/glut.h>
#include <stdio.h>

float PerspectiveView::projection [ 16 ];
float PerspectiveView::modelview [ 16 ];
int PerspectiveView::buttons;
int PerspectiveView::mouseX, PerspectiveView::mouseY;

void PerspectiveView::init ( void )
{
	buttons = 0;
	mouseX = 0;
	mouseY = 0;

	glMatrixMode(GL_PROJECTION);
	glPushMatrix ( );
		glLoadIdentity ( );
		gluPerspective(60, (float)( glutGet ( GLUT_WINDOW_WIDTH ) )/(float)( glutGet ( GLUT_WINDOW_HEIGHT ) ), 0.1f, 2000);	
		glGetFloatv ( GL_PROJECTION_MATRIX, projection );
	glPopMatrix ( );
	glMatrixMode ( GL_MODELVIEW );
	glPushMatrix ( );
		glLoadIdentity ( );
		glGetFloatv ( GL_MODELVIEW_MATRIX, modelview );
	glPopMatrix ( );
	modelview [ 14 ] = -150;
}

void PerspectiveView::mouseButton ( int button, int state, int x, int y )
{
	mouseX = x;
	mouseY = y;
	buttons = buttons + ( ( 1 - state - ( ( buttons >> button ) & 0x01 ) ) << button ); 
}

void PerspectiveView::mouseMove ( int x, int y )
{
	float rotMat [ 16 ];
	float newMat [ 16 ];
	int i;

	if ( buttons & 0x01 )
	{
		float dx, dy;

		dx = ( mouseX - x ) / (float)glutGet ( GLUT_WINDOW_WIDTH );
		dy = ( mouseY - y ) / (float)glutGet ( GLUT_WINDOW_HEIGHT );

		glPushMatrix ( );
			glLoadIdentity ( );
			glRotatef ( -360 * dy, 1, 0, 0 );
			glRotatef ( -360 * dx, 0, 1, 0 );
			glGetFloatv ( GL_MODELVIEW_MATRIX, rotMat );
		glPopMatrix ( );

		newMat [ 3 ] = 0;
		newMat [ 7 ] = 0;
		newMat [ 11 ] = 0;
		newMat [ 12 ] = 0;
		newMat [ 13 ] = 0;
		newMat [ 14 ] = 0;
		newMat [ 15 ] = 1;

		newMat [ 0 ] = modelview [ 0 ] * rotMat [ 0 ] + modelview [ 1 ] * rotMat [ 4 ] + modelview [ 2 ] * rotMat [ 8 ];
		newMat [ 1 ] = modelview [ 0 ] * rotMat [ 1 ] + modelview [ 1 ] * rotMat [ 5 ] + modelview [ 2 ] * rotMat [ 9 ];
		newMat [ 2 ] = modelview [ 0 ] * rotMat [ 2 ] + modelview [ 1 ] * rotMat [ 6 ] + modelview [ 2 ] * rotMat [ 10 ];
		newMat [ 4 ] = modelview [ 4 ] * rotMat [ 0 ] + modelview [ 5 ] * rotMat [ 4 ] + modelview [ 6 ] * rotMat [ 8 ];
		newMat [ 5 ] = modelview [ 4 ] * rotMat [ 1 ] + modelview [ 5 ] * rotMat [ 5 ] + modelview [ 6 ] * rotMat [ 9 ];
		newMat [ 6 ] = modelview [ 4 ] * rotMat [ 2 ] + modelview [ 5 ] * rotMat [ 6 ] + modelview [ 6 ] * rotMat [ 10 ];
		newMat [ 8 ] = modelview [ 8 ] * rotMat [ 0 ] + modelview [ 9 ] * rotMat [ 4 ] + modelview [ 10 ] * rotMat [ 8 ];
		newMat [ 9 ] = modelview [ 8 ] * rotMat [ 1 ] + modelview [ 9 ] * rotMat [ 5 ] + modelview [ 10 ] * rotMat [ 9 ];
		newMat [ 10 ] = modelview [ 8 ] * rotMat [ 2 ] + modelview [ 9 ] * rotMat [ 6 ] + modelview [ 10 ] * rotMat [ 10 ];

		for ( i = 0; i < 11; i++ )
		{
			modelview [ i ] = newMat [ i ];
		}
	}
	else
	{
		if ( buttons & 0x04 )
		{
			float dx, dy;

			dx = ( mouseX - x ) / (float)glutGet ( GLUT_WINDOW_WIDTH );
			dy = ( mouseY - y ) / (float)glutGet ( GLUT_WINDOW_HEIGHT );

			modelview [ 12 ] += dx * modelview [ 14 ];
			modelview [ 13 ] -= dy * modelview [ 14 ];
		}
		else
		{
			if ( buttons & 0x02 )
			{
				float dy;

				dy = ( mouseY - y ) / (float)glutGet ( GLUT_WINDOW_HEIGHT );

				modelview [ 14 ] -= dy * 100;

				if ( modelview [ 14 ] < -1000 )
				{
					modelview [ 14 ] = -1000;
				}
			}
		}
	}
	mouseX = x;
	mouseY = y;
}

void PerspectiveView::reshape ( int width, int height )
{
	printf ( "called\n" );
	glViewport ( 0, 0, width, height );
	glMatrixMode(GL_PROJECTION);
	glPushMatrix ( );
		glLoadIdentity ( );
		gluPerspective(60, (float)( width )/(float)( height ), 0.1f, 2000);	
		glGetFloatv ( GL_PROJECTION_MATRIX, projection );
	glPopMatrix ( );
	glMatrixMode ( GL_MODELVIEW );
}
