#ifndef VEC3_H
#define VEC3_H

#include <stdio.h>
#include <math.h>

class Vec3
{
protected:
	float pos [ 3 ];

public:
	Vec3 ( float x = 0, float y = 0, float z = 0 )
	{
		pos [ 0 ] = x;
		pos [ 1 ] = y;
		pos [ 2 ] = z;
	}

	Vec3 ( float *vec )
	{
		memcpy ( pos, vec, 3 * sizeof ( float ) );
	}

	float &operator[] ( int ind )
	{
		assert ( ind >=0 && ind < 3 );
		return pos [ ind ];
	}

	float operator[] ( int ind ) const
	{
		assert ( ind >=0 && ind < 3 );
		return pos [ ind ];
	}

	Vec3 operator += ( const Vec3 &vec )
	{
		pos [ 0 ] += vec [ 0 ];
		pos [ 1 ] += vec [ 1 ];
		pos [ 2 ] += vec [ 2 ];

		return *this;
	}

	Vec3 operator -= ( const Vec3 &vec )
	{
		pos [ 0 ] -= vec [ 0 ];
		pos [ 1 ] -= vec [ 1 ];
		pos [ 2 ] -= vec [ 2 ];

		return *this;
	}

	Vec3 operator + ( Vec3 &vec ) const
	{
		Vec3 rvalue = *this;

		rvalue += vec;

		return rvalue;
	}

	Vec3 operator - ( Vec3 &vec ) const
	{
		Vec3 rvalue = *this;

		rvalue -= vec;

		return rvalue;
	}

	Vec3 operator *= ( float s )
	{
		pos [ 0 ] *= s;
		pos [ 1 ] *= s;
		pos [ 2 ] *= s;

		return *this;
	}

	Vec3 operator /= ( float s )
	{
		*this *= ( 1 / s );

		return *this;
	}

	Vec3 operator * ( float s ) const
	{
		Vec3 rvalue = *this;

		rvalue *= s;

		return rvalue;
	}

	Vec3 operator / ( float s ) const
	{
		Vec3 rvalue = *this;

		rvalue *= (1 / s);

		return rvalue;
	}

	static Vec3 crossProd ( const Vec3 &vec1, const Vec3 &vec2 )
	{
		Vec3 rvalue;

		rvalue [ 0 ] = vec1 [ 1 ] * vec2 [ 2 ] - vec1 [ 2 ] * vec2 [ 1 ];
		rvalue [ 1 ] = vec1 [ 2 ] * vec2 [ 0 ] - vec1 [ 0 ] * vec2 [ 2 ];
		rvalue [ 2 ] = vec1 [ 0 ] * vec2 [ 1 ] - vec1 [ 1 ] * vec2 [ 0 ];

		return rvalue;
	}

	static float dotProd ( const Vec3 &vec1, const Vec3 &vec2 )
	{
		return vec1 [ 0 ] * vec2 [ 0 ] + vec1 [ 1 ] * vec2 [ 1 ] + vec1 [ 2 ] * vec2 [ 2 ];
	}

	float lenSq ( void ) const
	{
		return dotProd ( *this, *this );
	}

	float len ( void ) const
	{
		return sqrt ( lenSq ( ) );
	}

	void normalize ( void )
	{
		*this /= len ( );
	}
};

#endif