#include <windows.h>
#include <GL/glut.h>
#include <gl/gl.h>
#include <stdio.h>
#include "Mesh.h"
#include "PerspectiveView.h"
#include <time.h>

Mesh *mesh;
clock_t start, finish;
int frames = 0;
int type = 0;
char filename [ 256 ];

void display(void)
{
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

	glMatrixMode ( GL_PROJECTION );
	glLoadMatrixf ( PerspectiveView::projection );
	glMatrixMode ( GL_MODELVIEW );
	glLoadMatrixf ( PerspectiveView::modelview );

	// Draw polys in grey
	glColor3f(0.75, 0.75, 0.75);
	mesh->drawMesh ( type );
	glFlush ( );
	frames++;

	finish = clock ( );

	if ( ( finish - start ) / (double)CLOCKS_PER_SEC > 1 )
	{
		system ( "cls" );
		printf ( "%d vertices / %d faces\n", mesh->getNumVerts ( ), mesh->getNumTris ( ) );
		printf ( "%s\n", mesh->drawNames [ type ] );
		printf ( "fps: %f\n", frames * CLOCKS_PER_SEC / (float)( finish - start ) );
		start = finish;
		frames = 0;
	}

	glutSwapBuffers();
}

void init(void)
{
	glClearColor(0.0, 0.0, 0.0, 0.0);
	glShadeModel(GL_SMOOTH);
	glEnable(GL_DEPTH_TEST);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef ( 0, 0, -2 );

	glEnable(GL_LIGHTING);
	glLightf(GL_LIGHT0, GL_POSITION, (2.0, 2.0, 1.0));
	glLightf(GL_LIGHT0, GL_DIFFUSE, (0.5, 0.5, 0.5, 1.0));
	glLightf(GL_LIGHT0, GL_AMBIENT, (0.5, 0.5, 0.5, 1.0));
	glEnable(GL_LIGHT0);
	glEnable ( GL_CULL_FACE );
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);

	mesh = Mesh::loadObj ( filename );
	printf ( "Model loaded with %d verts and %d triangles\n", mesh->getNumVerts ( ), mesh->getNumTris ( ) );
	
	Vec3 minCoord, maxCoord;
	mesh->calcBoundingBox ( minCoord, maxCoord );
	Vec3 midpoint = ( minCoord + maxCoord ) / -2;
	mesh->translateSelf ( midpoint );
	midpoint = maxCoord - minCoord;
	if ( midpoint [ 1 ] > midpoint [ 0 ] )
	{
		midpoint [ 0 ] = midpoint [ 0 ];
	}
	if ( midpoint [ 2 ] > midpoint [ 0 ] )
	{
		midpoint [ 0 ] = midpoint [ 2 ];
	}
	mesh->scaleSelf ( 130 / midpoint [ 0 ] );

	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);
	
	start = clock ( );
//	mesh->initVBOs ( );
}

void keyboard ( unsigned char key, int x, int y )
{
	switch (key)
	{
	case '6':
		mesh->initVBOs ( );
	case '5':
	case '4':
	case '3':
	case '2':
	case '1':
		type = key - '1';
		break;
	}
}

void idle ( void )
{
	glutPostRedisplay ( );
}

int main(int argc, char** argv)
{
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE|GLUT_RGB|GLUT_DEPTH);
	glutInitWindowSize(500, 500);
	glutInitWindowPosition(100, 100);
	glutCreateWindow("Polygon Pusher");
	
	if ( argc > 1 )
	{
		strcpy ( filename, argv [ 1 ] );
	}
	else
	{
		strcpy ( filename, "data/cow.obj" );
	}

	init();
	PerspectiveView::init ( );

	glutDisplayFunc(display);
	glutReshapeFunc(&(PerspectiveView::reshape));
	glutMouseFunc(&(PerspectiveView::mouseButton));
	glutMotionFunc(&(PerspectiveView::mouseMove));
	glutKeyboardFunc(keyboard);
	glutIdleFunc ( idle );
	glutMainLoop();
	return 0;
}
